/*----------------------------------------------------------------------*/
/*                                                                      */
/*                      -------------------                             */
/*                                                                      */
/*                             LIGO                                     */
/*                                                                      */
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.      */
/*                                                                      */
/*                     (C) The LIGO Project, 2012.                      */
/*                                                                      */
/*                                                                      */
/*----------------------------------------------------------------------*/

///	@file controllerIop.c
///	@brief Main scheduler program for compiled real-time kernal object. \n
/// 	@detail More information can be found in the following DCC document:
///<	<a
///< href="https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=7688">T0900607
///< CDS RT Sequencer Software</a>
///	@author R.Bork, A.Ivanov
///     @copyright Copyright (C) 2014 LIGO Project      \n
///<    California Institute of Technology              \n
///<    Massachusetts Institute of Technology           \n\n
///     @license This program is free software: you can redistribute it and/or
///     modify
///<    it under the terms of the GNU General Public License as published by
///<    the Free Software Foundation, version 3 of the License. \n This program
///<    is distributed in the hope that it will be useful, but WITHOUT ANY
///<    WARRANTY; without even the implied warranty of MERCHANTABILITY or
///<    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
///<    for more details.

#include "controllerko.h"

#ifdef DOLPHIN_TEST
#include "dolphin.c"
#endif

#if defined( XMIT_DOLPHIN_TIME ) || defined( USE_DOLPHIN_TIMING )
volatile TIMING_SIGNAL* pcieTimer;
#endif

// Duotone diags struct
duotone_diag_t dt_diag;

// Variables for setting IOP->APP I/O
int ioClockDac = DAC_PRELOAD_CNT;
// int ioMemCntr = 0;
int ioMemCntrDac = DAC_PRELOAD_CNT;
// DAC variable
int dacEnable = 0;
int dacWriteEnable =
    0; /// @param dacWriteEnable  No DAC outputs until >4 times through code
int        dacTimingErrorPending[ MAX_DAC_MODULES ];
static int dacTimingError = 0;
int        dacWatchDog = 0;

int pBits[ 9 ] = { 1, 2, 4, 8, 16, 32, 64, 128, 256 };

int getGpsTime( unsigned int* tsyncSec, unsigned int* tsyncUsec );

// Include C code modules
#include "moduleLoad.c"
#if defined( RUN_WO_IO_MODULES ) || defined( USE_DOLPHIN_TIMING )
#include "mapVirtual.c"
#include <drv/no_ioc_timing.c>
#include "verify_slots.c"
#include <drv/dac_info.c>
#include <drv/adc_info.c>

// dummy placeholder function that should never be called when there's no IOC
int iop_dac_preload() {return 0;}
#else
#include "map.c"
#include <drv/iop_adc_functions.c>
#include <drv/iop_dac_functions.c>
#include "sync21pps.c"
#include "verify_slots.c"
#include <drv/dac_info_ioc2.c>
#include <drv/adc_info_ioc2.c>
#endif


//***********************************************************************
// TASK: fe_start_controller()
//***********************************************************************
/// This function is the main real-time sequencer or scheduler for IOP kernel
/// module code built using the RCG. \n 
/// This code runs in a continuous loop at the rate specified in the RCG model.
/// The loop is synchronized and triggered by the arrival of ADC data, the ADC
/// module in turn is triggered to sample by the clock provided by the LIGO
/// Timing Distribution System.
///	-
void*
fe_start_controller( void* arg )
{
    int        ii, jj, kk, ll; // Dummy loop counter variables
    static int cpuClock[ CPU_TIMER_CNT ]; ///  @param cpuClock[] Code timing
                                          ///  diag variables

    volatile RFM_FE_COMMS* pEpicsComms; /// @param *pEpicsComms Pointer to EPICS
                                        /// shared memory space
    int status; /// @param status Typical function return value
    one_pps_sync_t onePps;
    onePps.time = IOP_IO_RATE * 4; /// @param onePpsTime One PPS diagnostic check
    unsigned long nanotime = 100000;
    int           sync21pps = 0; ///@param flag for code sync to 1PPS signal
// FOR DIAGNOSTIC TESTING ONLY
#ifdef DIAG_TEST
    loopbacktest_t loopback;
    int lastWDread[ MAX_DAC_MODULES ] ;
    memset( lastWDread, 0, sizeof( lastWDread ) );
#endif
    int        dcuId; /// @param dcuId DAQ ID number for this process
    int diagWord =
        0; /// @param diagWord Code diagnostic bit pattern returned to EPICS
    int system = 0; // Legacy: Allow same code to run muliple instances of a model
    int syncSource =
        SYNC_SRC_NONE; /// @param syncSource Code startup synchronization source
    int mxStat = 0; /// @param mxStat: Used to status local_dc comms
    int mxDiag = 0; ///@param mxDiag: Used to status local_dc comms
    int mxDiagR = 0; ///@param mxDiagR: Used to status local_dc comms

    int feStatus = 0; /// @param Word for storing IOP error flags
    int dkiTrip = 0;

    unsigned int  usec = 0;
    unsigned long cpc;
    float         duotoneTimeDac;
    float         duotoneTime;

    int        usloop = 1;  // Number of upsamle loops when running ADC faster
                            // than code
    int        hkp_cycle = 0;
    double     adcval[ MAX_ADC_MODULES ][ MAX_ADC_CHN_PER_MOD ];
    volatile u32        dacWDread = 0;
    int        dacnum = 0;
    adcInfo_t* padcinfo;
    // Normally preload DAC FIFOs on startup, but can be bypassed
#ifdef NO_DAC_PRELOAD
    int dac_preload = 0;
    // DAC FIFO check will produce DAC error only for FIFO full
    int dac_fault_armed = 0;
#else
    int dac_preload = 1;
    // DAC FIFO check will produce DAC error for all FIFO errors
    int dac_fault_armed = 1;
#endif

    /// **********************************************************************************************\n
    /// Start Initialization Process \n
    /// **********************************************************************************************\n

    /// \> Flush L1 cache
    memset( fp, 0, 64 * 1024 );
    memset( fp, 1, 64 * 1024 );
    clflush_cache_range( (void*)fp, 64 * 1024 );

    fz_daz( ); /// \> Kill the denorms!

    /// \> Init comms with EPICS process */
    pEpicsComms = (volatile RFM_FE_COMMS*)_epics_shm;
    pLocalEpics = (volatile CDS_EPICS*)&pEpicsComms->epicsSpace;
    pEpicsDaq = (volatile char*)&( pLocalEpics->epicsOutput );
    padcinfo = (adcInfo_t*)&adcinfo;
#ifdef OVERSAMPLE
    /// \> Zero out filter histories
    memset( dHistory, 0, sizeof( dHistory ) );
    memset( dDacHistory, 0, sizeof( dDacHistory ) );
#endif

    /// \> Set pointers to filter module data buffers. \n
    /// - ---- Prior to V2.8, separate local/shared memories for FILT_MOD
    /// data.\n
    /// - ---- V2.8 and later, code uses EPICS shared memory only. This was done
    /// to: \n
    /// - -------- Allow daqLib.c to retrieve filter module data directly from
    /// shared memory. \n
    /// - -------- Avoid copy of filter module data between to memory locations,
    /// which was slow. \n
    pDsp[ system ] = (FILT_MOD*)( &pEpicsComms->dspSpace );
    pCoeff[ system ] = (VME_COEF*)( &pEpicsComms->coeffSpace );
    dspPtr[ system ] = (FILT_MOD*)( &pEpicsComms->dspSpace );

    /// \> Clear the FE reset which comes from Epics
    pLocalEpics->epicsInput.vmeReset = 0;

    // Clear input masks
    pLocalEpics->epicsInput.burtRestore_mask = 0;
    pLocalEpics->epicsInput.dacDuoSet_mask = 0;

    /// < Read in all Filter Module EPICS coeff settings
    for ( ii = 0; ii < MAX_MODULES; ii++ )
    {
        checkFiltReset( ii,
                        dspPtr[ 0 ],
                        pDsp[ 0 ],
                        &dspCoeff[ 0 ],
                        MAX_MODULES,
                        pCoeff[ 0 ] );
    }

    // Need this FE dcuId to make connection to FB
    dcuId = pLocalEpics->epicsInput.dcuId;
    pLocalEpics->epicsOutput.dcuId = dcuId;

    /// \> Init IIR filter banks
    //   Initialize filter banks  *********************************************
    for ( system = 0; system < NUM_SYSTEMS; system++ )
    {
        for ( ii = 0; ii < MAX_MODULES; ii++ )
        {
            for ( jj = 0; jj < FILTERS; jj++ )
            {
                for ( kk = 0; kk < MAX_COEFFS; kk++ )
                {
                    dspCoeff[ system ].coeffs[ ii ].filtCoeff[ jj ][ kk ] = 0.0;
                }
                dspCoeff[ system ].coeffs[ ii ].filtSections[ jj ] = 0;
            }
        }
    }

    /// \> Initialize all filter module excitation signals to zero
    for ( system = 0; system < NUM_SYSTEMS; system++ )
        for ( ii = 0; ii < MAX_MODULES; ii++ )
            pDsp[ 0 ]->data[ ii ].exciteInput = 0.0;

    /// \> Initialize IIR filter bank values
    if ( initVars( pDsp[ 0 ], pDsp[ 0 ], dspCoeff, MAX_MODULES, pCoeff[ 0 ] ) )
    {
        pLocalEpics->epicsOutput.fe_status = FILT_INIT_ERROR;
        fe_status_return = FILT_INIT_ERROR;
        return 0;
    }

    udelay( 1000 );

/// \> Initialize DAQ variable/software
#if !defined( NO_DAQ ) && !defined( IOP_TASK )
    /// - ---- Set data range limits for daqLib routine
    daq.filtExMin = GDS_16K_EXC_MIN;
    daq.filtTpMin = GDS_16K_TP_MIN;
    daq.filtExMax = daq.filtExMin + MAX_MODULES;
    daq.filtExSize = MAX_MODULES;
    daq.xExMin = daq.filtExMax;
    daq.xExMax = daq.xExMin + GDS_MAX_NFM_EXC;
    daq.filtTpMax = daq.filtTpMin + MAX_MODULES * TP_PER_FM;
    daq.filtTpSize = MAX_MODULES * TP_PER_FM;
    daq.xTpMin = daq.filtTpMax;
    daq.xTpMax = daq.xTpMin + GDS_MAX_NFM_TP;

    /// - ---- Initialize DAQ function
    status = daqWrite( 0,
                       dcuId,
                       daq,
                       DAQ_RATE,
                       testpoint,
                       dspPtr[ 0 ],
                       0,
                       (int*)( pLocalEpics->epicsOutput.gdsMon ),
                       xExc,
                       pEpicsDaq );
    if ( status == -1 )
    {
        pLocalEpics->epicsOutput.fe_status = DAQ_INIT_ERROR;
        fe_status_return = DAQ_INIT_ERROR;
        vmeDone = 1;
        return ( 0 );
    }

#endif

    /// - ---- Assign DAC testpoint pointers
    for ( ii = 0; ii < cdsPciModules.dacCount; ii++ )
        for ( jj = 0; jj < MAX_DAC_CHN_PER_MOD;
              jj++ ) // 16 per DAC regardless of the actual
            testpoint[ MAX_DAC_CHN_PER_MOD * ii + jj ] =
                floatDacOut + MAX_DAC_CHN_PER_MOD * ii + jj;

    // Zero out storage
    memset( floatDacOut, 0, sizeof( floatDacOut ) );

    pLocalEpics->epicsOutput.ipcStat = 0;
    pLocalEpics->epicsOutput.fbNetStat = 0;
    pLocalEpics->epicsOutput.tpCnt = 0;

    // Clear the code exit flag
    vmeDone = 0;
    fe_status_return = 0;

    /// \> Call user model defined application software initialization routine.
    iopDacEnable = feCode( cycleNum,
                           adcval,
                           dacOut,
                           dspPtr[ 0 ],
                           &dspCoeff[ 0 ],
                           (struct CDS_EPICS*)pLocalEpics,
                           FE_CODE_INIT );

    // Initialize timing info variables
    initializeTimingDiags( &timeinfo );

    // Initialize duotone measurement signals
    initializeDuotoneDiags( &dt_diag );

    /// \> Initialize the ADC modules *************************************
    pLocalEpics->epicsOutput.fe_status = INIT_ADC_MODS;
    status = iop_adc_init( padcinfo );

    /// \> Initialize the DAC module variables
    /// **********************************
    pLocalEpics->epicsOutput.fe_status = INIT_DAC_MODS;

    status = iop_dac_init( dacTimingErrorPending );

    pLocalEpics->epicsOutput.fe_status = INIT_SYNC;

    // Set ADC read timeout to 2sec (in iop_adc_functions.c)
    // Need to wait longer on startup for ADC data as it can
    // take up to 1.25 second for timing receiver to sync to
    // 1 second mark and start the ADC/DAC clocks.
    adcinfo.startup_sync = 1;

    // Reset timing diagnostics
    pLocalEpics->epicsOutput.diagWord = 0;
    pLocalEpics->epicsOutput.timeDiag = 0;

    /// \> Determine code startup synchronization and timing source.
    // if Contec 1616 BIO present, Gen1 TDS receiver will be used for timing.
    if ( cdsPciModules.card_count[ CON_1616DIO ] )
        syncSource = SYNC_SRC_TDS;
    // If PCIe timing card found (IOC Gen2), use that for startup sync
    else if ( cdsPciModules.card_count[ LPTC ] > 0 )
        syncSource = SYNC_SRC_LTC;
    // If not on aligo timing system, will try to sync to 1PPS signal on ADC
    // This is legacy support for non-LIGO sites that still use an eLIGO
    // timing system.
    else
        syncSource = SYNC_SRC_1PPS;

// Other options based on Parameter block settings
#ifdef XMIT_DOLPHIN_TIME
    // This is an IOP that transmits time and cycle count over Dolphin net
    pcieTimer =
        (volatile TIMING_SIGNAL*)( (volatile char*)( cdsPciModules
                                                         .dolphinWrite[ 0 ] ) +
                                   IPC_PCIE_TIME_OFFSET );
#endif
#ifdef USE_DOLPHIN_TIMING
    // This is an IOP w/o an I/O chassis, runs by receiving time/cycle over
    // Dolphin net.
    syncSource = SYNC_SRC_DOLPHIN;
    pcieTimer =
        (volatile TIMING_SIGNAL*)( (volatile char*)( cdsPciModules
                                                         .dolphinRead[ 0 ] ) +
                                   IPC_PCIE_TIME_OFFSET );
#endif

#ifdef NO_SYNC
    // Model parameter block set to no_sync=1
    // Code will use an unsynchronized external clock to run
    syncSource = SYNC_SRC_NONE;
#endif

#ifdef RUN_WO_IO_MODULES
    syncSource = SYNC_SRC_TIMER;
#endif

#ifdef USE_ADC_CLOCK
    // Cymac option to run with clock supplied by an ADC card internal timer.
    syncSource = SYNC_SRC_INTERNAL;
#endif

#ifdef TEST_1PPS
    // Test use only.
    // This allows testing of sync to 1PPS on system which has normal LIGO
    // timing system.
    switch ( syncSource )
    {
    // Need to start TDS clocks for testing
    case SYNC_SRC_TDS:
        start_tds_clocks( tdsCount );
        break;
    // If test system has LIGO PCIe timing card, then start clocks
    case SYNC_SRC_LTC:
        // Set status to slot checking
        // pLocalEpics->epicsOutput.fe_status = IOC_SLOT_CHK;
        // Card map to bp slot verification
        lptc_enable_all_slots( &cdsPciModules );
        lptc_start_clock( &cdsPciModules );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        udelay( MAX_UDELAY );
        lptc_status_update( &cdsPciModules );
        break;
    default:
        break;
    }
    udelay( MAX_UDELAY );
    udelay( MAX_UDELAY );
    syncSource = SYNC_SRC_1PPS;
#endif
    // Send sync source info to epics
    pLocalEpics->epicsOutput.timeErr = syncSource;

    // Start the ADC/DAC clocks based on syncSource
    switch ( syncSource )
    {
    // For Gen2 IOC, use PCIe timing card to setup and sync timing
    case SYNC_SRC_LTC:
        // Standard IOC startup
        lptc_enable_all_slots( &cdsPciModules );
        // Stop ADC/DAC clocks
        lptc_stop_clock( &cdsPciModules );
        // Enable all ADC cards
        gscEnableAdcModule ( &cdsPciModules,GSAI_ALL_CARDS,GSAI_DMA_MODE );
        // Enable all DAC cards 
        gscEnableDacModule ( &cdsPciModules,GSAI_ALL_CARDS,DAC_CLK_ENABLE );
        // Preload DAC FIFOs for timing diags
        if ( dac_preload )
            iop_dac_preload( GSAI_ALL_CARDS );
        // Start ADC/DAC clocks
        lptc_start_clock( &cdsPciModules );
        // Get GPS time
        gps_receiver_locked =
            lptc_get_gps_time( &cdsPciModules, &timeSec, &usec );
        // Subtract a second because first cycle will increment time by one
        timeSec--;
        break;
    // For Gen1 IOC, use Contec1616 module as interface to timing receiver
    // card to setup and sync timing
    case SYNC_SRC_TDS:
        // Stop ADC/DAC clocks
        stop_tds_clocks( tdsCount );
        // Enable all ADC/DAC cards
        gscEnableAdcModule ( &cdsPciModules,GSAI_ALL_CARDS,GSAI_DMA_MODE );
        gscEnableDacModule ( &cdsPciModules,GSAI_ALL_CARDS,DAC_CLK_ENABLE );
        // Preload DAC FIFOs for timing diags
        if ( dac_preload )
            iop_dac_preload( GSAI_ALL_CARDS );
        // Start ADC/DAC clocks
        start_tds_clocks( tdsCount );
        timeSec = current_time_fe( ) - 1;
        if ( cdsPciModules.gpsType == TSYNC_RCVR )
        {
            timeSec = getGpsSecTsync( );
            gps_receiver_locked = getGpsuSecTsync( &usec );
            // Subtract a second because first cycle will increment by one
            timeSec--;
        }
        break;
    // For legacy systems and Cymacs
    case SYNC_SRC_1PPS:
    case SYNC_SRC_NONE:
        // Do Not Preload values to DAC FIFOs
        dac_fault_armed = 0;
        syncSource = sync2pps_signal( &onePps, padcinfo, cpuClock );
        // Already have the cycle 0 data, so setting sync21pps
        // will have the main loop skip adc read on first cycle
        sync21pps = 1;
        // Need to set onePps vars for use in 1PPS timing diag later
        onePps.signalHi = 1;
        onePps.time = cycleNum;
        pLocalEpics->epicsOutput.timeErr = syncSource;
        timeSec = current_time_fe( ) - 1;
        break;
#ifdef USE_DOLPHIN_TIMING
    // For FE computer without an I/O chassis, use timing info from
    // Dolphin network
    case SYNC_SRC_DOLPHIN:
        timeSec = sync2master( pcieTimer );
        if ( timeSec == -1 )
        {
            timeSec = current_time_fe( ) - 1;
        }
        break;
#endif
    case SYNC_SRC_TIMER:
        timeSec = sync2cpuclock( );
        break;
    case SYNC_SRC_INTERNAL:
        gscEnableAdcModule ( &cdsPciModules,GSAI_ALL_CARDS,GSAI_DMA_MODE );
        gscEnableDacModule ( &cdsPciModules,GSAI_ALL_CARDS,DAC_CLK_ENABLE );
        if ( dac_preload )
            iop_dac_preload( GSAI_ALL_CARDS );
        timeSec = current_time_fe( ) - 1;
        break;
    default:
        dac_preload = 0;
        // Do Not Preload values to DAC FIFOs
        dac_fault_armed = 0;
        gscEnableDacModule ( &cdsPciModules,GSAI_ALL_CARDS,DAC_CLK_ENABLE );
        gscEnableAdcModule ( &cdsPciModules,GSAI_ALL_CARDS,GSAI_DMA_MODE );
        timeSec = current_time_fe( ) - 1;
        break;
    }

    pLocalEpics->epicsOutput.fe_status = NORMAL_RUN;
    pLocalEpics->epicsOutput.timeErr = syncSource;

#ifdef REMOTE_GPS
    timeSec = remote_time( (struct CDS_EPICS*)pLocalEpics );
#endif

#ifdef ADD_SEC_ON_START
    // This is a special for LHO PEMMID
    // Can remove once LHO moves FE compujters out of mids
    timeSec++;
#endif

    // Get CPU clock time for diags
    adcinfo.adcTime = rdtsc_ordered( );

    /// ******************************************************************************\n
    /// Enter the infinite FE control loop
    /// ******************************************\n
    /// ******************************************************************************\n
    // Calculate how many CPU cycles per code cycle
    cpc = cpu_khz * 1000;
    cpc /= CYCLE_PER_SECOND;
#ifdef TEST_1PPS
    syncSource = SYNC_SRC_LTC;
#endif

    // Run forever
#ifdef NO_CPU_SHUTDOWN  // Run without core lock
    while ( !kthread_should_stop( ) && !vmeDone )
    {
#else
    while ( !vmeDone )
    { 
#endif
        // NORMAL OPERATION -- Wait for ADC data ready
        // *****************************************************************************************
        /// \> On 1PPS mark \n
        if ( cycleNum == 0 )
        {
            /// - ---- Check awgtpman status.
            pLocalEpics->epicsOutput.awgStat =
                ( pEpicsComms->padSpace.awgtpman_gps != timeSec );
            if ( pLocalEpics->epicsOutput.awgStat )
                feStatus |= FE_ERROR_AWG;
            /// - ---- Check if DAC outputs are enabled, report error.
            if ( !iopDacEnable || dkiTrip )
                feStatus |= FE_ERROR_DAC_ENABLE;

            /// - ---- Increment GPS second
            timeSec++;
            pLocalEpics->epicsOutput.timeDiag = timeSec;
            // Capture code startup time
            if ( cycle_gps_time == 0 )
            {
                timeinfo.startGpsTime = timeSec;
                pLocalEpics->epicsOutput.startgpstime = timeinfo.startGpsTime;
            }
            cycle_gps_time = timeSec;
        }
#ifdef NO_CPU_SHUTDOWN
#ifndef RUN_WO_IO_MODULES
        // If core not locked, need occasional delays to keep Linux happy
        if ( ( cycleNum % 2048 ) == 0 )
        {
            usleep_range( 1, 3 );
        }
#endif
#endif

        // **********************************************************************
        // Read ADC data
        // **********************************************************************
        // If sync to 1pps, first cycle 0 data hase already been read by the
        // sync21pps routine, so don't read again.
        if ( !sync21pps )
            status = iop_adc_read( padcinfo, cpuClock );
        // Clear sync21pps so reads now continue every cycle
        sync21pps = 0;
        adcinfo.startup_sync = 0;
        // Check for an ADC timeout error return
        // This indicates one or more ADC modules in not clocking properly
        if ( status == ADC_TO_ERROR )
        {
            feStatus |= adc_status_update( &adcinfo );
            pLocalEpics->epicsOutput.stateWord = FE_ERROR_ADC;
            pLocalEpics->epicsOutput.diagWord |= ADC_TIMEOUT_ERR;
            pLocalEpics->epicsOutput.fe_status = ADC_TO_ERROR;
            pLocalEpics->epicsOutput.epicsSync++;
            stop_working_threads = 1;
            vmeDone = 1;
            continue;
        }
        // If ADC channels not where they should be, we have no option
        // but to exit from the RT code ie loops would be working with
        // wrong input data.
        if ( status == CHAN_HOP_ERROR )
        {
            feStatus |= adc_status_update( &adcinfo );
            pLocalEpics->epicsOutput.stateWord = FE_ERROR_ADC;
            pLocalEpics->epicsOutput.fe_status = CHAN_HOP_ERROR;
            stop_working_threads = 1;
            vmeDone = 1;
            pLocalEpics->epicsOutput.epicsSync++;
            continue;
        }

        // Following captures time after ADC reads
        cpuClock[ CPU_TIME_USR_START ] = rdtsc_ordered( );

        // In normal operation, the following for loop runs only once per IOP
        // code cycle. This for loop runs > once per cycle if ADC is clocking
        // faster then IOP is running ie
        // specifically added for fast 18bit ADC running at 128KS/sec and
        // greater.
        for ( usloop = 0; usloop < UNDERSAMPLE; usloop++ )
        {
            // Move ADC data from read buffer to local buffer for passing to
            // user code.
            for ( ii = 0; ii < cdsPciModules.adcCount; ii++ )
            {
                for ( jj = 0; jj < cdsPciModules.adcChannels[ ii ]; jj++ )
                {
                    adcval[ ii ][ jj ] = dWord[ ii ][ jj ][ usloop ];
                }
            }
            // *****************************************************************
            /// \> Call the front end specific application  ******************\n
            /// - -- This is where the user application produced by RCG gets
            /// called and executed. \n\n
            //
            iopDacEnable = feCode( cycleNum,
                                   adcval,
                                   dacOut,
                                   dspPtr[ 0 ],
                                   &dspCoeff[ 0 ],
                                   (struct CDS_EPICS*)pLocalEpics,
                                   FE_CODE_PROCESS );
            cpuClock[ CPU_TIME_USR_END ] = rdtsc_ordered( );
            // ******************************************************************
            if ( usloop == 0 )
            {
                // ********************************************************************
                /// WRITE DAC OUTPUTS *****************************************
                /// \n
                // ********************************************************************
#ifndef DAC_WD_OVERRIDE
                // If a DAC module has bad timing then quit writing outputs
                // COMMENT OUT NEXT LINE FOR TEST STAND w/bad DAC cards.
                if ( dacTimingError )
                    iopDacEnable = 0;
#endif
                // Write out data to DAC modules
                dkiTrip = iop_dac_write( );
            }
            odcStateWord = 0;

            // ***********************************************************************
            /// BEGIN HOUSEKEEPING
            /// ************************************************ \n
            // ***********************************************************************

            pLocalEpics->epicsOutput.cycle = cycleNum;
            // *****************************************************************
            /// \> Cycle 0: \n
            /// - ---- Read IRIGB time if symmetricom card (this is not
            /// standard, but supported for earlier cards. \n
            // *****************************************************************
            if ( cycleNum == HKP_READ_SYMCOM_IRIGB &&
                 ( syncSource == SYNC_SRC_TDS || syncSource == SYNC_SRC_LTC ) )
            {
                if ( cdsPciModules.gpsType == SYMCOM_RCVR )
                {
                    // Retrieve time set in at adc read and report offset from
                    // 1PPS
                    gps_receiver_locked = getGpsTime( &timeSec, &usec );
                    pLocalEpics->epicsOutput.irigbTime = usec;
                    if ( ( usec > MAX_IRIGB_SKEW || usec < MIN_IRIGB_SKEW ) &&
                         cdsPciModules.gpsType != 0 )
                    {
                        diagWord |= TIME_ERR_IRIGB;
                        feStatus |= FE_ERROR_TIMING;
                    }
                }
                /// - ---- Calc duotone diagnostic mean values for past second
                /// and reset.
                dt_diag.meanAdc = dt_diag.totalAdc / CYCLE_PER_SECOND;
                dt_diag.totalAdc = 0.0;
                dt_diag.meanDac = dt_diag.totalDac / CYCLE_PER_SECOND;
                dt_diag.totalDac = 0.0;
            }

            // *****************************************************************
            /// \> Cycle 1: Get cycle 0 time offset from  1 sec mark in usec.
            /// If Spectricom IRIGB, get IRIG-B usec information.
            /// If LIGO PCIe timing rcvr, get timing card usec information.
            /// In both cases, usec register should be held by having read
            /// the seconds register on cycle 0.
            // *****************************************************************
            if ( cycleNum == HKP_READ_TSYNC_IRIBB &&
                 ( syncSource == SYNC_SRC_TDS || syncSource == SYNC_SRC_LTC ) )
            {
                if ( cdsPciModules.gpsType == TSYNC_RCVR )
                    gps_receiver_locked = getGpsuSecTsync( &usec );
                if ( syncSource == SYNC_SRC_LTC )
                    usec = lptc_get_gps_usec( &cdsPciModules ) - 15;
                pLocalEpics->epicsOutput.irigbTime = usec;
                /// - ---- If not in acceptable range, generate error.
                if(syncSource == SYNC_SRC_LTC)
                {
                    if ( ( usec > MAX_IRIGB_SKEW ) ||
                         ( usec < MIN_IRIGB_SKEW_LTC ) )
                    {
                        feStatus |= FE_ERROR_TIMING;
                        diagWord |= TIME_ERR_IRIGB;
                    }
                }
                else
                {
                    if ( ( usec > MAX_IRIGB_SKEW ) ||
                         ( usec < MIN_IRIGB_SKEW ) )
                    {
                        feStatus |= FE_ERROR_TIMING;
                        diagWord |= TIME_ERR_IRIGB;
                    }
                }
            }

#ifdef SERVO64K
            /// \> Update duotone diag information
            if ( syncSource == SYNC_SRC_TDS ||
                 syncSource == SYNC_SRC_LTC )
            {
                dt_diag.dac[ ( cycleNum + DT_SAMPLE_OFFSET ) % IOP_IO_RATE ] =
                    dWord[ ADC_DUOTONE_BRD ][ DAC_DUOTONE_CHAN ][ usloop ];
                dt_diag.totalDac +=
                    dWord[ ADC_DUOTONE_BRD ][ DAC_DUOTONE_CHAN ][ usloop ];
                dt_diag.adc[ ( cycleNum + DT_SAMPLE_OFFSET ) % IOP_IO_RATE ] =
                    dWord[ ADC_DUOTONE_BRD ][ ADC_DUOTONE_CHAN ][ usloop ];
                dt_diag.totalAdc +=
                    dWord[ ADC_DUOTONE_BRD ][ ADC_DUOTONE_CHAN ][ usloop ];
            }
#else
            /// \> Update duotone diag information
            if ( syncSource == SYNC_SRC_TDS ||
                 syncSource == SYNC_SRC_LTC && usloop == 0 )
            {
                dt_diag.dac[ ( hkp_cycle + DT_SAMPLE_OFFSET ) % IOP_IO_RATE ] =
                    dWord[ ADC_DUOTONE_BRD ][ DAC_DUOTONE_CHAN ][ usloop ];
                dt_diag.totalDac +=
                    dWord[ ADC_DUOTONE_BRD ][ DAC_DUOTONE_CHAN ][ usloop ];
                dt_diag.adc[ ( hkp_cycle + DT_SAMPLE_OFFSET ) % IOP_IO_RATE ] =
                    dWord[ ADC_DUOTONE_BRD ][ ADC_DUOTONE_CHAN ][ usloop ];
                dt_diag.totalAdc +=
                    dWord[ ADC_DUOTONE_BRD ][ ADC_DUOTONE_CHAN ][ usloop ];
            }
#endif

            // *****************************************************************
            /// \> Cycle 16, perform duotone diag calcs.
            // *****************************************************************
            if ( hkp_cycle == HKP_DT_CALC &&
                 ( syncSource == SYNC_SRC_TDS || syncSource == SYNC_SRC_LTC ) )
            {
                duotoneTime =
                    duotime( DT_SAMPLE_CNT, dt_diag.meanAdc, dt_diag.adc );
                pLocalEpics->epicsOutput.dtTime = duotoneTime;
                if ( ( ( duotoneTime < MIN_DT_DIAG_VAL ) ||
                       ( duotoneTime > MAX_DT_DIAG_VAL ) ) )
                    feStatus |= FE_ERROR_TIMING;
                duotoneTimeDac =
                    duotime( DT_SAMPLE_CNT, dt_diag.meanDac, dt_diag.dac );
                pLocalEpics->epicsOutput.dacDtTime = duotoneTimeDac;
            }

            // *****************************************************************
            /// \> Cycle 17, set/reset DAC duotone switch if request has
            /// changed.
            // *****************************************************************
            if ( cycleNum == HKP_DAC_DT_SWITCH && syncSource == SYNC_SRC_TDS )
            {
                if ( dt_diag.dacDuoEnable != pLocalEpics->epicsInput.dacDuoSet )
                {
                    dt_diag.dacDuoEnable = pLocalEpics->epicsInput.dacDuoSet;
                    if ( dt_diag.dacDuoEnable )
                        CDIO1616Output[ 0 ] = TDS_START_ADC_NEG_DAC_POS;
                    else
                        CDIO1616Output[ 0 ] =
                            TDS_START_ADC_NEG_DAC_POS | TDS_NO_DAC_DUOTONE;
                    CDIO1616Input[ 0 ] = contec1616WriteOutputRegister(
                        &cdsPciModules, tdsControl[ 0 ], CDIO1616Output[ 0 ] );
                }
            }
            if ( cycleNum == HKP_DAC_DT_SWITCH && syncSource == SYNC_SRC_LTC )
            {
                if ( dt_diag.dacDuoEnable != pLocalEpics->epicsInput.dacDuoSet )
                {
                    dt_diag.dacDuoEnable = pLocalEpics->epicsInput.dacDuoSet;
                    lptc_dac_duotone( &cdsPciModules, dt_diag.dacDuoEnable );
                }
            }

            // *****************************************************************
            /// \> Cycle 18, Send timing info to EPICS at 1Hz
            // *****************************************************************
            if ( cycleNum == HKP_TIMING_UPDATES )
            {
                sendTimingDiags2Epics( pLocalEpics, &timeinfo, &adcinfo );

                pLocalEpics->epicsOutput.dacEnable = dacEnable;

                if ( ( adcinfo.adcHoldTime > CYCLE_TIME_ALRM_HI ) ||
                     ( adcinfo.adcHoldTime < CYCLE_TIME_ALRM_LO ) )
                {
                    diagWord |= FE_ADC_HOLD_ERR;
                    feStatus |= FE_ERROR_TIMING;
                }
                if ( timeinfo.cycleTimeMax > CYCLE_TIME_ALRM_HI )
                {
                    diagWord |= FE_PROC_TIME_ERR;
                    feStatus |= FE_ERROR_TIMING;
                }
                pLocalEpics->epicsOutput.stateWord = feStatus;
                feStatus = 0;
                if ( pLocalEpics->epicsInput.diagReset || initialDiagReset )
                {
                    initialDiagReset = 0;
                    pLocalEpics->epicsInput.diagReset = 0;
                    pLocalEpics->epicsInput.ipcDiagReset = 1;
                    timeinfo.cycleTimeMax = 0;
                    diagWord = 0;
                    ipcErrBits = 0;
                    for ( jj = 0; jj < cdsPciModules.adcCount; jj++ )
                        adcinfo.adcRdTimeMax[ jj ] = 0;
                }
                pLocalEpics->epicsOutput.diagWord = diagWord;
            }

            // *****************************************************************
            /// \> Check for requests for filter module clear history requests.
            /// This is spread out over a number of cycles.
            // Spread out filter coeff update, but keep updates at 16 Hz
            // here we are rounding up:
            //   x/y rounded up equals (x + y - 1) / y
            //
            // *****************************************************************
            {
                static const unsigned int mpc =
                    ( MAX_MODULES + ( FE_RATE / 16 ) - 1 ) /
                    ( FE_RATE / 16 ); // Modules per cycle
                unsigned int smpc = mpc * subcycle; // Start module counter
                unsigned int empc = smpc + mpc; // End module counter
                unsigned int i;
                for ( i = smpc; i < MAX_MODULES && i < empc; i++ )
                    checkFiltReset( i,
                                    dspPtr[ 0 ],
                                    pDsp[ 0 ],
                                    &dspCoeff[ 0 ],
                                    MAX_MODULES,
                                    pCoeff[ 0 ] );
            }

            // *****************************************************************
            /// \> Check if code exit is requested
            if ( cycleNum == MAX_MODULES )
                vmeDone = stop_working_threads |
                    checkEpicsReset( cycleNum, (struct CDS_EPICS*)pLocalEpics );

            // *****************************************************************
            // If synced to 1PPS on startup, continue to check that code
            // is still in sync with 1PPS.
            // This is NOT normal aLIGO mode, but still used at AEI and ANU.
#ifdef TEST_1PPS
            if ( syncSource == SYNC_SRC_LTC )
#else
            if ( syncSource == SYNC_SRC_1PPS )
#endif
            {
                // Check if front end continues to be in sync with 1pps
                // If not, set sync error flag
                sync21pps_check( &onePps, padcinfo );
                if ( onePps.time > 1 )
                {
                    diagWord |= TIME_ERR_1PPS;
                    feStatus |= FE_ERROR_TIMING;
                }
            }

// FOR DIAGNOSTIC TESTING ONLEY
// Following is only used on automated test system
// These signals are looped back from other test models
// to measure phase delays
#ifdef DIAG_TEST
            for ( ii = 0; ii < 15; ii++ )
            {
                if ( ii < 5 )
                    loopback.value = adcinfo.adcData[ 0 ][ ii ];
                if ( ( ii > 4 ) && ( ii < 10 ) )
                    loopback.value = adcinfo.adcData[ 1 ][ ii - 5 ];
                if ( ii > 9 )
                    loopback.value = adcinfo.adcData[ 0 ][ ( ii - 2 ) ];
                if ( ( loopback.value > 400 ) && ( loopback.signalHi[ ii ] == 0 ) )
                {
                    loopback.time[ ii ] = cycleNum;
                    loopback.signalHi[ ii ] = 1;
                    if ( ( ii % 5 ) == 0 )
                        pLocalEpics->epicsOutput.timingTest[ ii ] =
                            (cycleNum * 15.26 ) / UNDERSAMPLE;
                    // Control apps do not see 1pps until after IOP signal loops
                    // around and back into ADC channel 0, therefore, need to
                    // subtract IOP loop time.
                    else
                        pLocalEpics->epicsOutput.timingTest[ ii ] =
                            ( cycleNum * 15.26 /UNDERSAMPLE) -
                            pLocalEpics->epicsOutput.timingTest[ 0 ];
                }
                // Reset the diagnostic for next cycle
                if ( cycleNum > 2000 )
                    loopback.signalHi[ ii ] = 0;
            }
#endif

// *****************************************************************
/// \>  Write data to DAQ.
// *****************************************************************
#ifndef NO_DAQ

            // Call daqLib
            pLocalEpics->epicsOutput.daqByteCnt =
                daqWrite( 1,
                          dcuId,
                          daq,
                          DAQ_RATE,
                          testpoint,
                          dspPtr[ 0 ],
                          0,
                          (int*)( pLocalEpics->epicsOutput.gdsMon ),
                          xExc,
                          pEpicsDaq );
            // Send the current DAQ block size to the awgtpman for TP number
            // checking
            pEpicsComms->padSpace.feDaqBlockSize = curDaqBlockSize;
            pLocalEpics->epicsOutput.tpCnt = tpPtr->count & 0xff;
            feStatus |= ( FE_ERROR_EXC_SET & tpPtr->count );
            if ( FE_ERROR_EXC_SET & tpPtr->count )
                odcStateWord |= ODC_EXC_SET;
            else
                odcStateWord &= ~( ODC_EXC_SET );
            if ( pLocalEpics->epicsOutput.daqByteCnt > DAQ_DCU_RATE_WARNING )
                feStatus |= FE_ERROR_DAQ;
#endif

            // *****************************************************************
            /// \> Cycle 19, write updated diag info to EPICS
            // *****************************************************************
            if ( cycleNum == HKP_DIAG_UPDATES )
            {
                pLocalEpics->epicsOutput.ipcStat = ipcErrBits;
                if ( ipcErrBits & 0xf )
                    feStatus |= FE_ERROR_IPC;
                // Create FB status word for return to EPICS
                mxStat = 0;
                mxDiagR = daqPtr->reqAck;
                if ( ( mxDiag & 1 ) != ( mxDiagR & 1 ) )
                    mxStat = 1;
                if ( ( mxDiag & 2 ) != ( mxDiagR & 2 ) )
                    mxStat += 2;
                pLocalEpics->epicsOutput.fbNetStat = mxStat;
                mxDiag = mxDiagR;
                if ( mxStat != MX_OK )
                    feStatus |= FE_ERROR_DAQ;
                ;
                if ( pLocalEpics->epicsInput.overflowReset )
                {
                    if ( pLocalEpics->epicsInput.overflowReset )
                    {
                        for ( ii = 0; ii < 16; ii++ )
                        {
                            for ( jj = 0; jj < cdsPciModules.adcCount; jj++ )
                            {
                                adcinfo.overflowAdc[ jj ][ ii ] = 0;
                                adcinfo.overflowAdc[ jj ][ ii + 16 ] = 0;
                                pLocalEpics->epicsOutput
                                    .overflowAdcAcc[ jj ][ ii ] = 0;
                                pLocalEpics->epicsOutput
                                    .overflowAdcAcc[ jj ][ ii + 16 ] = 0;
                            }
                            for ( jj = 0; jj < cdsPciModules.dacCount; jj++ )
                            {
                                pLocalEpics->epicsOutput
                                    .overflowDacAcc[ jj ][ ii ] = 0;
                            }
                        }
                    }
                }
                if ( ( pLocalEpics->epicsInput.overflowReset ) ||
                     ( overflowAcc > OVERFLOW_CNTR_LIMIT ) )
                {
                    pLocalEpics->epicsInput.overflowReset = 0;
                    pLocalEpics->epicsOutput.ovAccum = 0;
                    overflowAcc = 0;
                }
            }

            // *****************************************************************
            /// \> Cycle 20, Update latest DAC output values to EPICS
            // *****************************************************************
            if ( subcycle == HKP_DAC_EPICS_UPDATES )
            {
                // Send DAC output values at 16Hzfb
                for ( jj = 0; jj < cdsPciModules.dacCount; jj++ )
                {
                    for ( ii = 0; ii < MAX_DAC_CHN_PER_MOD; ii++ )
                    {
                        pLocalEpics->epicsOutput.dacValue[ jj ][ ii ] =
                            dacOutEpics[ jj ][ ii ];
                    }
                }
            }

            // *****************************************************************
            /// \> Cycle 21, Update ADC/DAC status to EPICS.
            // *****************************************************************
            if ( cycleNum == HKP_ADC_DAC_STAT_UPDATES )
            {
                pLocalEpics->epicsOutput.ovAccum = overflowAcc;
                feStatus |= adc_status_update( &adcinfo );
                feStatus |= dac_status_update( &dacinfo );
                // pLocalEpics->epicsOutput.fe_status = NORMAL_RUN;
            }

#if !defined( RUN_WO_IO_MODULES ) && !defined( USE_DOLPHIN_TIMING )
            // Following read I/O
            // In systems > 64KHz, do not want to read multiple times
            // within an ADC read period.
            if ( usloop == 0 )
            {
                // *****************************************************************
                /// \> Cycle 400 to 400 + numDacModules, write DAC heartbeat to
                /// AI chassis (only for 18 and 20bit DAC modules), via DAC DIO port.
                // Only one write per code cycle to reduce cycle time
                // Lower 4 bits of 8 bit DAC DIO set as outputs.
                // *****************************************************************
                if ( hkp_cycle >= HKP_DAC_WD_CLK &&
                     hkp_cycle < ( HKP_DAC_WD_CLK + cdsPciModules.dacCount ) )
                {
                    dacnum = hkp_cycle - HKP_DAC_WD_CLK;
                    // AI WD only used with 18 and 20bit DAC modules
                    if ( cdsPciModules.dacType[ dacnum ] == GSC_18AO8 ||
                         cdsPciModules.dacType[ dacnum ] == GSC_20AO8 )
                    {
                        // Only send to DAC if IOP is writing normally
                        if ( iopDacEnable && !dacChanErr[ dacnum ] )
                            _dacPtr[ dacnum ]->CSR_DIO =
                                ( (timeSec & 1) | DAC_DIO_L4B_OUTPUT );
                    }
                }

                // *****************************************************************
                /// \> Cycle 500 to 500 + numDacModules, read back watchdog from
                /// AI chassis (18 and 20bit DAC only)
                // Check once per second on code cycle HKP_DAC_WD_CHK to DAC
                // count Only one read per code cycle to reduce cycle time
                // Upper 4 bits of 8 bit DAC DIO set as inputs.
                // *****************************************************************
                if ( hkp_cycle >= HKP_DAC_WD_CHK &&
                     hkp_cycle < ( HKP_DAC_WD_CHK + cdsPciModules.dacCount ) )
                {
                    dacnum = hkp_cycle - HKP_DAC_WD_CHK;
                    if ( cdsPciModules.dacType[ dacnum ] == GSC_18AO8 ||
                         cdsPciModules.dacType[ dacnum ] == GSC_20AO8 )
                    {
                        // Read the DAC DIO register
                        dacWDread = _dacPtr[ dacnum ]->CSR_DIO;
#ifdef DIAG_TEST
                        if( (dacWDread & 1) == lastWDread[ dacnum ] )
                            pLocalEpics->epicsOutput.statDac[ dacnum ] &=
                                ~( DAC_WD_BIT );
                        else
                            pLocalEpics->epicsOutput.statDac[ dacnum ] |=
                                DAC_WD_BIT;
                        lastWDread[ dacnum ] = (dacWDread & 1);
#else
                        // Bit 8 of DAC DIO is AI chassis enable status
                        // This bit should be high if AIC enabled
                        if ( ( ( dacWDread >> 8 ) & 1 ) > 0 )
                        {
                            pLocalEpics->epicsOutput.statDac[ dacnum ] &=
                                ~( DAC_WD_BIT );
                        }
                        else
                            pLocalEpics->epicsOutput.statDac[ dacnum ] |=
                                DAC_WD_BIT;
#endif
                    }
                }

                // *****************************************************************
                /// \> Cycle 600 to 600 + numDacModules, Check DAC FIFO Sizes to
                /// determine if DAC modules are synched to code
                /// - ---- 18bit DAC reads out FIFO size dirrectly, but 16bit
                /// module only has a 4 bit register area for FIFO empty,
                /// quarter full, etc. So, to make these bits useful in 16 bit
                /// module, code must set a proper FIFO size in map.c code.
                // This code runs once per second.
                // *****************************************************************
                if ( hkp_cycle >= HKP_DAC_FIFO_CHK &&
                     hkp_cycle < ( HKP_DAC_FIFO_CHK + cdsPciModules.dacCount ) )
                {
                    dacnum = hkp_cycle - HKP_DAC_FIFO_CHK;
                    status = check_dac_buffers( dacnum, dac_fault_armed );
                    if ( dacTimingError )
                        feStatus |= FE_ERROR_DAC;
                }
            }

            // Check LIGO timing card and backplane status

            if ( usloop == 0 && syncSource == SYNC_SRC_LTC )
            {
                if ( hkp_cycle >= HKP_IOC_SLOT_STATUS &&
                     hkp_cycle < ( HKP_IOC_SLOT_STATUS + LPTC_IOC_SLOTS ) )
                {
                    jj = hkp_cycle - HKP_IOC_SLOT_STATUS;
                    pLocalEpics->epicsOutput.lptMon[ jj ] =
                        lptc_get_slot_status( &cdsPciModules, jj );
                }

                if ( hkp_cycle >= HKP_IOC_SLOT_CONFIG &&
                     hkp_cycle < ( HKP_IOC_SLOT_CONFIG + LPTC_IOC_SLOTS ) )
                {
                    jj = hkp_cycle - HKP_IOC_SLOT_CONFIG;
                    pLocalEpics->epicsOutput.lptMon[ ( jj + LPTC_IOC_SLOTS ) ] =
                        lptc_get_slot_config( &cdsPciModules, jj );
                }

                if ( hkp_cycle == HKP_IOC_BP_CONFIG )
                {
                    pLocalEpics->epicsOutput.lpt_bp_config =
                        lptc_get_bp_config( &cdsPciModules );
                }

                if ( hkp_cycle == HKP_IOC_BP_STATUS )
                {
                    pLocalEpics->epicsOutput.lpt_bp_status =
                        lptc_get_bp_status( &cdsPciModules );
                }

                if ( hkp_cycle == HKP_IOC_LPTC_STATUS )
                {
                    pLocalEpics->epicsOutput.lpt_status =
                        lptc_get_lptc_status( &cdsPciModules );
                }
            }
#endif

            // *****************************************************************
            // Update end of cycle information
            // *****************************************************************
            if ( usloop == UNDERSAMPLE-1 )
            {
                // Capture end of cycle time.
                cpuClock[ CPU_TIME_CYCLE_END ] = rdtsc_ordered( );

                /// \> Compute code cycle time diag information.
                timeinfo.cycleTime = ( cpuClock[ CPU_TIME_CYCLE_END ] -
                                       cpuClock[ CPU_TIME_CYCLE_START ] ) /
                    CPURATE;

                captureEocTiming(
                    cycleNum, cycle_gps_time, &timeinfo, &adcinfo );

                adcinfo.adcHoldTime =
                    ( cpuClock[ CPU_TIME_CYCLE_START ] - adcinfo.adcTime ) /
                    CPURATE;
                adcinfo.adcTime = cpuClock[ CPU_TIME_CYCLE_START ];
                // Calc the max time of one IOP cycle.  Each IOP cycle is the
                // same duration as UNDERSAMPLE model cycles.
                // For IOP, more interested in time to get thru ADC read code
                // and send to control apps
                timeinfo.usrTime = ( cpuClock[ CPU_TIME_USR_START ] -
                                     cpuClock[ CPU_TIME_CYCLE_START ] ) /
                    CPURATE;
                if ( timeinfo.usrTime > timeinfo.usrTimeMax )
                    timeinfo.usrTimeMax = timeinfo.usrTime;
            }

            /// \> Update internal cycle counters
            cycleNum += 1;

// FOR DIAGNOSTIC TESTING ONLEY
// Introduces clocking errors for verification that code will catch
// and report these errors.
#ifdef DIAG_TEST
            if ( pLocalEpics->epicsInput.bumpCycle != 0 )
            {
                cycleNum += pLocalEpics->epicsInput.bumpCycle;
                pLocalEpics->epicsInput.bumpCycle = 0;
            }
#endif
            cycleNum %= CYCLE_PER_SECOND;
            if ( subcycle == DAQ_CYCLE_CHANGE )
            {
                daqCycle = ( daqCycle + 1 ) % DAQ_NUM_DATA_BLOCKS_PER_SECOND;
                if ( !( daqCycle % 2 ) )
                    pLocalEpics->epicsOutput.epicsSync = daqCycle;
            }
            if ( subcycle ==
                 END_OF_DAQ_BLOCK ) /*we have reached the 16Hz second barrier*/
            {
                /* Reset the data cycle counter */
                subcycle = 0;
            }
            else
            {
                /* Increment the internal cycle counter */
                subcycle++;
            }
        }
        if ( cycleNum == 0 )
            hkp_cycle = 0;
        else
            hkp_cycle++;

        /// \> If not exit request, then continue INFINITE LOOP  *******
    }

    pLocalEpics->epicsOutput.cpuMeter = 0;

    /* System reset command received */
    return (void*)-1;
}
